#! /usr/bin/env python3

from collections import OrderedDict, UserDict
# simplejson serializes namedtuples into objects and offers C-based acceleration. (Yay!)
import simplejson as json

from ..utils import odict


class Message(UserDict):
    '''
    A dictionary-like container for API messages.
    '''

    def __init__(self, *args, **kwargs):
        self.data = odict(*args)
        self.data.update(kwargs)

    @staticmethod
    def _from_odict(od):
        assert isinstance(od, OrderedDict)
        m = Message()
        m.data = od
        return m

    def encode(self):
        s = json.dumps(self.data, ensure_ascii=False)
        return s.encode('utf8', errors='replace')

    @staticmethod
    def decode(s):
        if isinstance(s, bytes):
            s = s.decode('utf8')
        root_od = json.loads(s, object_pairs_hook=OrderedDict)
        return Message._from_odict(root_od)



#! /usr/bin/env python3
import asyncio
import base64
from collections import OrderedDict
from itertools import chain
import numbers
import socket
import uuid

import aiohttp
from async_timeout import timeout as _timeout


def odict(*args):
    '''
    A short-hand for the constructor of OrderedDict.
    :code:`odict(('a':1), ('b':2))` is equivalent to
    :code:`OrderedDict([('a':1), ('b':2)])`.
    '''
    return OrderedDict(args)


def dict2kvlist(o):
    '''
    Serializes a dict-like object into a generator of the flatten list of
    repeating key-value pairs.  It is useful when using HMSET method in Redis.

    Example:
    >>> list(dict2kvlist({'a': 1, 'b': 2}))
    ['a', 1, 'b', 2]
    '''
    return chain.from_iterable((k, v) for k, v in o.items())


def generate_uuid():
    u = uuid.uuid4()
    # Strip the last two padding characters because u always has fixed length.
    return base64.urlsafe_b64encode(u.bytes)[:-2].decode('ascii')


def nmget(o, key_path, def_val=None, path_delimiter='.'):
    '''
    A short-hand for retrieving a value from nested mappings
    ("nested-mapping-get").  At each level it checks if the given "path"
    component in the given key exists and return the default value whenever
    fails.

    Example:
    >>> o = {'a':{'b':1}}
    >>> nmget(o, 'a', 0)
    {'b': 1}
    >>> nmget(o, 'a.b', 0)
    1
    >>> nmget(o, 'a/b', 0, '/')
    1
    >>> nmget(o, 'a.c', 0)
    0
    '''
    pieces = key_path.split(path_delimiter)
    while pieces:
        p = pieces.pop(0)
        if o is None or p not in o:
            return def_val
        o = o[p]
    return o


def readable_size_to_bytes(expr):
    if isinstance(expr, numbers.Real):
        return int(expr)
    elif isinstance(expr, str):
        try:
            v = float(expr)
            return int(v)
        except ValueError:
            suffix = expr[-1]
            suffix_map = {
                't': 1024 * 1024 * 1024 * 1024,
                'T': 1024 * 1024 * 1024 * 1024,
                'g': 1024 * 1024 * 1024,
                'G': 1024 * 1024 * 1024,
                'm': 1024 * 1024,
                'M': 1024 * 1024,
                'k': 1024,
                'K': 1024,
            }
            return int(float(expr[:-1]) * suffix_map[suffix])
    else:
        raise ValueError('unconvertible type')


async def curl(url, default_value=None, loop=None, timeout=0.2):
    try:
        async with aiohttp.ClientSession(loop=loop) as sess:
            with _timeout(timeout):
                async with sess.get(url) as resp:
                    assert resp.status == 200
                    body = await resp.text()
                    return body.strip()
    except (asyncio.TimeoutError, aiohttp.errors.ClientOSError, AssertionError) as e:
        if callable(default_value):
            return default_value()
        return default_value


async def get_instance_id(loop=None):
    return (await curl('http://169.254.169.254/latest/meta-data/instance-id',
                       lambda: 'i-{}'.format(socket.gethostname()), loop=loop))


async def get_instance_ip(loop=None):
    return (await curl('http://169.254.169.254/latest/meta-data/local-ipv4',
                       '127.0.0.1', loop=loop))


async def get_instance_type(loop=None):
    return (await curl('http://169.254.169.254/latest/meta-data/instance-type',
                       'unknown', loop=loop))


class AsyncBarrier:
    '''
    This class provides a simplified asyncio-version of threading.Barrier class.
    '''

    num_parties = 1
    loop = None
    cond = None

    def __init__(self, num_parties, loop=None):
        self.num_parties = num_parties
        self.count = 0
        self.loop = loop if loop else asyncio.get_event_loop()
        self.cond = asyncio.Condition(loop=self.loop)

    async def wait(self):
        async with self.cond:
            self.count += 1
            if self.count == self.num_parties:
                self.cond.notify_all()
            else:
                while self.count < self.num_parties:
                    await self.cond.wait()

    def reset(self):
        self.count = 0
        # FIXME: if there are waiting coroutines, let them
        #        raise BrokenBarrierError like threading.Barrier



#! /usr/bin/env python3

import enum


class SornaResponseTypes(enum.IntEnum):
    PONG          = 0
    SUCCESS       = 200
    INVALID_INPUT = 400
    FAILURE       = 500


class ManagerRequestTypes(enum.IntEnum):
    PING          = 0  # used for service status monitoring
    HEARTBEAT     = 1  # (unused)
    GET_OR_CREATE = 2
    DESTROY       = 3


class AgentRequestTypes(enum.IntEnum):
    PING           = 0  # used for service status monitoring
    EXECUTE        = 1
    CREATE_KERNEL  = 2
    DESTROY_KERNEL = 3
    RESTART_KERNEL = 4



'''
This module defines a series of Sorna-specific errors
based on HTTP Error classes from aiohttp.
Raising a SornaError automatically is automatically
mapped to a corresponding HTTP error with RFC7807-style
JSON-encoded description in its response body.
'''

from aiohttp import web
import simplejson as json

from .utils import odict


class SornaError(web.HTTPError):
    status_code = 500
    error_type  = 'https://api.sorna.io/probs/general-error'
    error_title = 'General Sorna API Error.'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.args = (self.status_code, self.reason, self.error_type)
        self.empty_body = False
        self.body = json.dumps({
            'type': self.error_type,
            'title': self.error_title,
        }).encode()

    def serialize(self):
        return odict(('type', self.error_type),
                     ('title', self.error_title))


class InvalidAuthParameters(web.HTTPBadRequest, SornaError):
    error_type  = 'https://api.sorna.io/probs/invalid-auth-params'
    error_title = 'Missing or invalid authorization parameters.'


class AuthorizationFailed(web.HTTPUnauthorized, SornaError):
    error_type  = 'https://api.sorna.io/probs/auth-failed'
    error_title = 'Credential/signature mismatch.'


class InstanceNotAvailable(web.HTTPServiceUnavailable, SornaError):
    error_type  = 'https://api.sorna.io/probs/instance-not-available'
    error_title = 'There is no available instance.'


class InstanceNotFound(web.HTTPNotFound, SornaError):
    error_type  = 'https://api.sorna.io/probs/instance-not-found'
    error_title = 'No such instance.'


class KernelNotFound(web.HTTPNotFound, SornaError):
    error_type  = 'https://api.sorna.io/probs/kernel-not-found'
    error_title = 'No such kernel.'


class QuotaExceeded(web.HTTPPreconditionFailed, SornaError):
    error_type  = 'https://api.sorna.io/probs/quota-exceeded'
    error_title = 'You have reached your resource limit.'


class RateLimitExceeded(web.HTTPTooManyRequests, SornaError):
    error_type  = 'https://api.sorna.io/probs/rate-limit-exceeded'
    error_title = 'You have reached your API query rate limit.'


class KernelCreationFailed(web.HTTPInternalServerError, SornaError):
    error_type  = 'https://api.sorna.io/probs/kernel-ceration-failed'
    error_title = 'Kernel creation has failed.'


class KernelDestructionFailed(web.HTTPInternalServerError, SornaError):
    error_type  = 'https://api.sorna.io/probs/kernel-destruction-failed'
    error_title = 'Kernel destruction has failed.'


class KernelExecutionFailed(web.HTTPInternalServerError, SornaError):
    error_type  = 'https://api.sorna.io/probs/kernel-execution-failed'
    error_title = 'Executing user code in the kernel has failed.'



#! /usr/bin/env python3

import argparse
import ipaddress
import pathlib


def port_no(s):
    try:
        port = int(s)
        assert port > 0
        assert port < 65536
    except (ValueError, AssertionError):
        msg = '{!r} is not a valid port number.'.format(s)
        raise argparse.ArgumentTypeError(msg)
    return port


def positive_int(s):
    try:
        val = int(s)
        assert val > 0
    except (ValueError, AssertionError):
        msg = '{!r} is not a positive integer.'.format(s)
        raise argparse.ArgumentTypeError(msg)
    return val


def host_port_pair(s):
    pieces = s.split(':', maxsplit=1)
    if len(pieces) == 1:
        msg = '{!r} should contain both IP address and port number.'.format(s)
        raise argparse.ArgumentTypeError(msg)
    elif len(pieces) == 2:
        ipaddr = pieces[0]
        try:
            port = int(pieces[1])
            assert port > 0
            assert port < 65536
        except (ValueError, AssertionError):
            msg = '{!r} is not a valid port number.'.format(s)
            raise argparse.ArgumentTypeError(msg)
    return (ipaddr, port)


def ipaddr(s):
    try:
        addr = ipaddress.ip_address(s)
    except ValueError as e:
        msg = '{!r} is not a valid IP address.'.format(s)
        raise argparse.ArgumentTypeError(msg)
    return addr


def path(val):
    if val is None:
        return None
    p = pathlib.Path(val)
    if not p.exists():
        msg = '{!r} is not a valid file/dir path.'.format(val)
        raise argparse.ArgumentTypeError(msg)
    return p
